package org.myorg;

import java.io.*;
import java.util.*;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.conf.*;
import org.apache.hadoop.io.*;
import org.apache.hadoop.mapred.*;
import org.apache.hadoop.util.*;
import org.apache.hadoop.mapreduce.Mapper.Context;


public class HadoopKernel extends Configured implements Tool {

    public native float[][] start(float[][]data, float eps, float stdv);


    public static class Map extends MapReduceBase implements Mapper<LongWritable, Text, NullWritable, FloatArrayWritable> {

      int dimension;
      double eps;
      double delta;
      double prob;
      float stdv;
      int tPts;
      int nPts;
      Random r;
      Vector<Text> values;


      public void configure(JobConf job) {
        dimension = job.getInt( "HadoopKernel.d", -1 );
        eps = (double) job.getFloat( "HadoopKernel.eps", -1 );
        delta = (double) job.getFloat( "HadoopKernel.delta", -1 );
        stdv = job.getFloat( "HadoopKernel.stdv", -1 );
        tPts = job.getInt( "HadoopKernel.tPts", -1 );
        double size = 1.0d / ( eps * eps );
        size *= Math.log( 1.0d / delta );
        nPts = (int) Math.ceil( size );

        prob = (double) nPts / tPts;
        r = new Random();

        System.out.println( "d: " + dimension );
        System.out.println( "eps: " + eps );
        System.out.println( "delta: " + delta );
        System.out.println( "tPts: " + tPts );
        System.out.println( "size: " + tPts );
        System.out.println( "nPts: " + nPts );
        System.out.println( "prob: " + prob );
      }


      public void map(LongWritable key, Text value, OutputCollector<NullWritable, FloatArrayWritable> output, Reporter reporter) throws IOException {
         if( tPts < 0 || eps < 0 || dimension < 0 || delta < 0 || prob < 0 || stdv < 0 ) throw new IOException( "param not correct" );

         if( r.nextDouble() < prob ) {
            String parts[] = value.toString().split(" ");
            if( parts.length < dimension ) return;

            FloatWritable[] fwa = new FloatWritable[dimension];
            for( int j = 0; j < fwa.length; ++j )
               fwa[j] = new FloatWritable( Float.valueOf( parts[j] ) );


            output.collect( NullWritable.get(), new FloatArrayWritable(fwa) );
         }

      }

    }

    public static class Reduce extends MapReduceBase implements Reducer<NullWritable, FloatArrayWritable, NullWritable, FloatArrayWritable> {

       float eps = 0;
       float stdv = 0;
       int dimension = 0;
       String libFName;
       OutputCollector<NullWritable,FloatArrayWritable> oc;

       Vector<float[]> points;
 
       public void configure( JobConf job ) {


          dimension = job.getInt( "HadoopKernel.d", -1 );
          eps = job.getFloat( "HadoopKernel.eps", -1 );
          stdv = job.getFloat( "HadoopKernel.stdv", -1 );
          points = new Vector<float[]>();

          System.out.println( "dimension: " + dimension );
          System.out.println( "eps: " + eps );
          System.out.println( "stdv: " + stdv );


          Path[] libFiles = new Path[0];
          try {
            libFiles = DistributedCache.getLocalCacheFiles(job);
          } catch (IOException ioe) {
            System.err.println("Caught exception while getting cached files: " + StringUtils.stringifyException(ioe));
          }
  
          for( Path pathF : libFiles ) {
            libFName = pathF.toString();
            System.out.println( libFName );
          }
        }

 
       public void reduce(NullWritable key, Iterator<FloatArrayWritable> values, OutputCollector<NullWritable, FloatArrayWritable> output, Reporter reporter) throws IOException {
          if( eps < 0 || dimension < 0 || stdv < 0 ) throw new IOException( "param not correct" );

          oc = output;

          while (values.hasNext()) {
            FloatWritable[] fwa = (FloatWritable[]) values.next().toArray();

            float[] currentPt = new float[dimension];
            for( int i = 0; i < dimension; ++i )
                 currentPt[i] = fwa[i].get();

            points.add( currentPt );

          }
       }


      public void close() {

        try {
           float[][] pointArr = new float[points.size()][dimension];
           for( int i = 0; i < points.size(); ++i )
              pointArr[i] = points.get(i);
  
           points = null; 
           System.load(libFName);
           HadoopKernel myLib = new HadoopKernel();
           float[][] sample = myLib.start(pointArr, eps, stdv);
           
           pointArr = null;
           for( int i = 0; i < sample.length; ++i ) {
   
              FloatWritable[] fwa = new FloatWritable[dimension];
              for( int j = 0; j < sample[i].length; ++j )
                 fwa[j] = new FloatWritable( sample[i][j] );
   
              oc.collect( NullWritable.get(), new FloatArrayWritable( fwa ) );
           } 
        } catch(Exception e) {}
      }
    }


    public int run(String[] args) throws Exception {

      if( args.length < 4 ) {
         System.out.println( "Usage: <infile> <outdir> -lib <libfile> -d <dimension> -delta <delta> -eps <epsilon> -stdv <standarddeviation> -tpts <totalpointsindataset>" );
         System.out.println( "Note: libfile must be in HDFS!!!" );
         return 1;
      }

      JobConf conf = new JobConf(getConf(), HadoopKernel.class);
      conf.setJobName("HadoopKernel");
      conf.set( "mapred.child.java.opts", "-server -Xmx4096m -Djava.net.preferIPv4Stack=true" );

      conf.setOutputKeyClass(NullWritable.class);
      conf.setOutputValueClass(FloatArrayWritable.class);

      conf.setMapperClass(Map.class);
      conf.setReducerClass(Reduce.class);
      conf.setNumReduceTasks(1);

      conf.setInputFormat(TextInputFormat.class);
      conf.setOutputFormat(TextOutputFormat.class);

      List<String> other_args = new ArrayList<String>();
      for (int i=0; i < args.length; ++i) {
        if ("-lib".equals(args[i])) {
          DistributedCache.addCacheFile(new Path(args[++i]).toUri(), conf);
        }
        if( "-d".equals(args[i]) ) {
          conf.setInt( "HadoopKernel.d", Integer.parseInt( args[++i] ) );
          System.out.println( "d: " + args[i] );
        }
        if( "-delta".equals(args[i]) ) {
          conf.setFloat( "HadoopKernel.delta", Float.parseFloat( args[++i] ) );
          System.out.println( "delta: " + args[i] );
        }
        if( "-eps".equals(args[i]) ) {
          conf.setFloat( "HadoopKernel.eps", Float.parseFloat( args[++i] ) );
          System.out.println( "eps: " + args[i] );
        }
        if( "-stdv".equals(args[i]) ) {
          conf.setFloat( "HadoopKernel.stdv", Float.parseFloat( args[++i] ) );
          System.out.println( "stdv: " + args[i] );
        }
        if( "-tpts".equals(args[i]) ) {
          conf.setInt( "HadoopKernel.tPts", Integer.parseInt( args[++i] ) );
          System.out.println( "tpts: " + args[i] );
        }
        else {
          other_args.add(args[i]);
        }
      }

      FileInputFormat.setInputPaths(conf, new Path(other_args.get(0)));
      FileOutputFormat.setOutputPath(conf, new Path(other_args.get(1)));
      RunningJob rj = JobClient.runJob(conf);

      return 0;
    }

    public static void main(String[] args) throws Exception {
      int res = ToolRunner.run(new Configuration(), new HadoopKernel(), args);
      System.exit(res);
    }
}
