package org.myorg;
import sort.Zorder;

import java.io.*;
import java.util.*;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.conf.*;
import org.apache.hadoop.io.*;
import org.apache.hadoop.mapred.*;
import org.apache.hadoop.util.*;
import org.apache.hadoop.mapreduce.Mapper.Context;


public class HadoopKernel extends Configured implements Tool {



    public static class Map extends MapReduceBase implements Mapper<LongWritable, Text, NullWritable, FloatArrayWritable> {

      int dimension;
      double eps;
      double delta;
      double prob;
      int tPts;
      int nPts;
      Random r;
      Vector<Text> values;


      public void configure(JobConf job) {
        dimension = job.getInt( "HadoopKernel.d", -1 );
        eps = (double) job.getFloat( "HadoopKernel.eps", -1 );
        delta = (double) job.getFloat( "HadoopKernel.delta", -1 );
        tPts = job.getInt( "HadoopKernel.tPts", -1 );
        double size = 1.0d / ( eps * eps );
        size *= Math.log( 1.0d / delta );
        nPts = (int) Math.ceil( size );

        prob = (double) nPts / tPts;
        r = new Random();

        System.out.println( "nPts: " + nPts );
      }


      public void map(LongWritable key, Text value, OutputCollector<NullWritable, FloatArrayWritable> output, Reporter reporter) throws IOException {
         if( eps < 0 || dimension < 0 ) throw new IOException( "param not correct" );

         if( r.nextDouble() < prob ) {
            String parts[] = value.toString().split(" ");
            if( parts.length < dimension ) return;

            FloatWritable[] fwa = new FloatWritable[dimension];
            for( int j = 0; j < fwa.length; ++j )
               fwa[j] = new FloatWritable( Float.valueOf( parts[j] ) );


            output.collect( NullWritable.get(), new FloatArrayWritable(fwa) );
         }

      }

    }

    public static class Reduce extends MapReduceBase implements Reducer<NullWritable, FloatArrayWritable, NullWritable, FloatArrayWritable> {

      int dimension;
      int k;
      int shift[];
      double eps;
      OutputCollector<NullWritable, FloatArrayWritable> oc;

      Vector<ZKernelRec> points;      

      public void configure(JobConf job) {
        dimension = job.getInt( "HadoopKernel.d", -1 );
        eps = job.getFloat( "HadoopKernel.eps", -1 );
        k = (int) Math.ceil( (4.0/eps) * Math.sqrt( Math.log( 4.0 / eps ) ) * Math.log( 4.0 / eps ) );


        System.out.println( "eps: " + eps );
        System.out.println( "k: " + k );
        System.out.println( "dim: " + dimension );

        shift = new int[dimension];
        points = new Vector<ZKernelRec>();

        // generate random shifts and avoid overflow
        Random r = new Random();
        for( int i = 0; i < shift.length; ++i )  {
           shift[i] = r.nextInt( Integer.MAX_VALUE - 500000 );
           System.out.println( "shift[" + i + "] = " + shift[i] );
        }
      }


      public void reduce(NullWritable key, Iterator<FloatArrayWritable> values, OutputCollector<NullWritable, FloatArrayWritable> output, Reporter reporter) throws IOException {
         if( dimension < 0 || k < 0 || eps < 0 ) throw new IOException( "param not correct" );

         oc = output;

         while( values.hasNext() ) {
            FloatWritable faw[] = (FloatWritable[]) values.next().toArray();        
    
            int[] coord = new int[dimension];
            float[] ocoord = new float[dimension];
            for( int i = 0; i < dimension; ++i ) {
                    float f = faw[i].get();
                    coord[i] = Zorder.toScaledInt( f );
                    ocoord[i] = f;
            }
            coord = Zorder.shift( coord, shift );
            String ZVal = Zorder.valueOf(dimension, coord);
   
            points.add( new ZKernelRec( ZVal, ocoord ) );
         }

      }

      public void close() {


         Collections.sort( points, new ZKernelRec.ZKernelRecComparator() );
         Vector<ZKernelRec> spoints = new Vector<ZKernelRec>();
         
         int skipLength = (int) Math.floor( ( (double) points.size() - (double)k ) / ( (double)k - 1.0 ) );
         if( skipLength == 0 ) skipLength = 1;
         System.out.println( "skip: " + skipLength );
         System.out.println( "k: " + k );
         System.out.println( "points: " + points.size() );

         for( int j = 0; j < points.size(); j += skipLength ) {
            FloatWritable farr[] = new FloatWritable[dimension];
            ZKernelRec zr = points.get(j);

            for( int i = 0; i < dimension; ++i )
               farr[i] = new FloatWritable( zr.coord[i] );

            try {
               oc.collect( NullWritable.get(), new FloatArrayWritable( farr ) ); 
            }
            catch( Exception e ) { System.out.println( e.toString() ); }

         }
      }
    }


    public int run(String[] args) throws Exception {

      if( args.length < 4 ) {
         System.out.println( "Usage: <infile> <outdir> -d <dimension> -delta <delta> -eps <epsilon> -tpts <totalpointsindataset>" );
         return 1;
      }

      JobConf conf = new JobConf(getConf(), HadoopKernel.class);
      conf.setJobName("HadoopKernel");

      conf.setOutputKeyClass(NullWritable.class);
      conf.setOutputValueClass(FloatArrayWritable.class);

      conf.setMapperClass(Map.class);
      conf.setReducerClass(Reduce.class);
      conf.setNumReduceTasks(1);

      conf.setInputFormat(TextInputFormat.class);
      conf.setOutputFormat(TextOutputFormat.class);

      List<String> other_args = new ArrayList<String>();
      for (int i=0; i < args.length; ++i) {
        if( "-d".equals(args[i]) ) {
          conf.setInt( "HadoopKernel.d", Integer.parseInt( args[++i] ) );
          System.out.println( "d: " + args[i] );
        }
        if( "-delta".equals(args[i]) ) {
          conf.setFloat( "HadoopKernel.delta", Float.parseFloat( args[++i] ) );
          System.out.println( "delta: " + args[i] );
        }
        if( "-eps".equals(args[i]) ) {
          conf.setFloat( "HadoopKernel.eps", Float.parseFloat( args[++i] ) );
          System.out.println( "eps: " + args[i] );
        }
        if( "-tpts".equals(args[i]) ) {
          conf.setInt( "HadoopKernel.tPts", Integer.parseInt( args[++i] ) );
          System.out.println( "tpts: " + args[i] );
        }
        else {
          other_args.add(args[i]);
        }
      }

      FileInputFormat.setInputPaths(conf, new Path(other_args.get(0)));
      FileOutputFormat.setOutputPath(conf, new Path(other_args.get(1)));
      RunningJob rj = JobClient.runJob(conf);

      return 0;
    }

    public static void main(String[] args) throws Exception {
      int res = ToolRunner.run(new Configuration(), new HadoopKernel(), args);
      System.exit(res);
    }
}
